from zodipy import emissivities
from zodipy import parameters
from zodipy._model import IPDModel, model_registry


model_registry.register_model(
    IPDModel(
        name="K98",
        parameters=parameters.K98,
    )
)

model_registry.register_model(
    IPDModel(
        name="Planck13",
        parameters=parameters.PLANCK,
        emissivities=emissivities.PLANCK_2013,
    )
)

model_registry.register_model(
    IPDModel(
        name="Planck15",
        parameters=parameters.PLANCK,
        emissivities=emissivities.PLANCK_2015,
    )
)

model_registry.register_model(
    IPDModel(
        name="Planck18",
        parameters=parameters.PLANCK,
        emissivities=emissivities.PLANCK_2018,
    )
)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict, Optional

from zodipy._components import Component
from zodipy._component_labels import ComponentLabel, LABEL_TO_CLASS
from zodipy._emissivities import Emissivity


class IPDModel:
    """An Interplanetary Dust Model.

    An IPDModel is a container for a unique combination of initialized Zodiacal
    Components and emissivity fits."""

    def __init__(
        self,
        name: str,
        parameters: Dict[ComponentLabel, Dict[str, float]],
        emissivities: Optional[Emissivity] = None,
    ) -> None:
        """Initializes an IPDModel given a set of parameters and emissivities."""

        self.name = name
        self.components: Dict[ComponentLabel, Component] = {}
        for component_label, component_parameters in parameters.items():
            component_class = LABEL_TO_CLASS[component_label]
            self.components[component_label] = component_class(**component_parameters)

        self.emissivities = emissivities

    @property
    def includes_earth_neighboring_components(self) -> bool:
        """Returns True if the model includes an Earth-neighboring component."""

        return (
            ComponentLabel.RING in self.components
            or ComponentLabel.FEATURE in self.components
        )

    def __getitem__(self, component_name: str) -> Component:
        """Returns a sky component from the cosmoglobe model."""

        return self.components[ComponentLabel(component_name)]


@dataclass
class IPDModelRegistry:
    """Container for registered IPDModels."""

    registry: Dict[str, IPDModel] = field(default_factory=dict)

    def register_model(self, model: IPDModel) -> None:
        """Adds a new IPD model to the registry."""

        if (name := model.name) in self.registry:
            raise ValueError(f"model by name {name} is already registered.")

        self.registry[name] = model

    def get_model(self, name: str) -> IPDModel:
        """Returns a IPDModel from the registry."""

        if name not in self.registry:
            raise ModuleNotFoundError(
                f"{name} is not a model in the registry. Avaliable models are "
                f"{', '.join(self.registry)}"
            )

        return self.registry[name]


model_registry = IPDModelRegistry()



from math import sin, cos
from typing import Tuple, Sequence, Union

from zodipy.models import model_registry
from zodipy._component_labels import ComponentLabel
import numpy as np
import astropy.constants as const


def get_component_density_grid(
    component: str,
    model: str = "K98",
    earth_coords: Sequence[float] = (0, 1, 0),
    xy_lim: int = 5,
    z_lim: int = 1,
    n: int = 200,
    return_grid: bool = False,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]:
    """Returns a 3D grid of the components density.

    Parameters
    ----------
    component
        The component for which we grid the density.
    model
        The Interplanetary Dust Model that contains the component.
    xy_lim
        xy-limit in AU.
    z_lim
        z-limit in AU.
    n
        Number of grid points in each axis.
    return_grid
        If True, then the meshgrid is returned alongside the density_grid.

    Returns
    -------
    density_grid, XX, YY, ZZ
        3D grid of the density of the component.
    """

    ipd_model = model_registry.get_model(model)
    component_class = ipd_model.components[ComponentLabel(component)]

    x_helio = np.linspace(-xy_lim, xy_lim, n)
    y_helio = x_helio.copy()
    z_helio = np.linspace(-z_lim, z_lim, n)

    x_prime = x_helio - component_class.x_0
    y_prime = y_helio - component_class.y_0
    z_prime = z_helio - component_class.z_0

    XX_helio, YY_helio, ZZ_helio = np.meshgrid(x_helio, y_helio, z_helio)
    XX_prime, YY_prime, ZZ_prime = np.meshgrid(x_prime, y_prime, z_prime)

    R_prime = np.sqrt(XX_prime ** 2 + YY_prime ** 2 + ZZ_prime ** 2)
    Z_prime = (
        XX_prime * sin(component_class.Ω) * sin(component_class.i)
        - YY_prime * cos(component_class.Ω) * sin(component_class.i)
        + ZZ_prime * cos(component_class.i)
    )

    X_earth_prime = earth_coords - component_class.X_component[0]
    θ_prime = np.arctan2(YY_prime, ZZ_prime) - np.arctan2(
        X_earth_prime[1], X_earth_prime[0]
    )

    density_grid = component_class.get_density(
        R_prime=R_prime, Z_prime=Z_prime, θ_prime=θ_prime
    )

    # The densities in the model is given in units of 1 / AU
    density_grid *= const.au.value

    if return_grid:
        return density_grid, XX_helio, YY_helio, ZZ_helio

    return density_grid



from typing import Protocol

import numpy as np


class EmissionFunction(Protocol):
    def __call__(
        self,
        distance_to_shell: np.ndarray,
        observer_position: np.ndarray,
        earth_position: np.ndarray,
        unit_vectors: np.ndarray,
        freq: float,
    ) -> np.ndarray:
        """Interface of the `get_emission` function of a Zodiacal Component."""


def line_of_sight_integrate(
    radial_distances: np.ndarray,
    get_emission_func: EmissionFunction,
    observer_position: np.ndarray,
    earth_position: np.ndarray,
    unit_vectors: np.ndarray,
    freq: float,
) -> np.ndarray:
    """Returns the integrated Zodiacal emission for a component (Trapezoidal).

    Parameters
    ----------
    radial_distances
        Array of discrete radial distances from the observer [AU].
    get_emission_func
        The `get_emission` function of the component.
    observer_position
        The heliocentric ecliptic cartesian position of the observer.
    earth_position
        The heliocentric ecliptic cartesian position of the Earth.
    unit_vectors
        Heliocentric ecliptic cartesian unit vectors pointing to each 
        position in space we that we consider.    
    freq
        Frequency at which to evaluate to Zodiacal Emission.

    Returns
    -------
    integrated_emission
        The line-of-sight integrated emission of the shells around an observer
        for a Zodiacal Component.
    """

    emission_previous = get_emission_func(
        distance_to_shell=radial_distances[0],
        observer_position=observer_position,
        earth_position=earth_position,
        unit_vectors=unit_vectors,
        freq=freq,
    )
    dR = np.diff(radial_distances)

    integrated_emission = np.zeros(unit_vectors.shape[-1])
    for r, dr in zip(radial_distances, dR):
        emission_current = get_emission_func(
            distance_to_shell=r,
            observer_position=observer_position,
            earth_position=earth_position,
            unit_vectors=unit_vectors,
            freq=freq,
        )
        integrated_emission += (emission_previous + emission_current) * dr / 2
        emission_previous = emission_current

    return integrated_emission



from typing import List

import healpy as hp
import numpy as np

from zodipy._components import Component
from zodipy._integration import line_of_sight_integrate


def instantaneous_emission(
    nside: int,
    freq: float,
    components: List[Component],
    emissivities: List[float],
    line_of_sights: List[np.ndarray],
    observer_positions: np.ndarray,
    earth_positions: np.ndarray,
    coord_out: str,
) -> np.ndarray:
    """Returns the simulated instantaneous emission.

    Parameters
    ----------
    nside
        HEALPIX map resolution parameter.
    freq
        Frequency at which to evaluate the Zodiacal emission [Hz].
    components
        List of Zodiacal Components for which to simulate the emission.
    emissivities
        Sequency of emissivities at the frequency given by `freq`, one for each
        component.
    observer_positions
        A sequence of (or a single) heliocentric cartesian positions of the 
        observer.
    earth_positions
        A sequence of (or a single) heliocentric cartesian positions of the 
        Earth.
    coord_out
        Coordinate frame of the output map.

    Returns
    -------
    emission
        Simulated (mean) instantaneous Zodiacal emission [MJy/sr].
    """

    npix = hp.nside2npix(nside)
    unit_vectors = _get_unit_vectors(
        nside=nside,
        pixels=np.arange(npix),
        coord_out=coord_out,
    )
    emission = np.zeros((len(components), npix))

    for observer_position, earth_position in zip(
        observer_positions,
        earth_positions,
    ):
        for idx, component in enumerate(components):
            integrated_comp_emission = line_of_sight_integrate(
                radial_distances=line_of_sights[idx],
                get_emission_func=component.get_emission,
                observer_position=observer_position,
                earth_position=earth_position,
                unit_vectors=unit_vectors,
                freq=freq,
            )

            emission[idx] += integrated_comp_emission * emissivities[idx]

    return emission * 1e20


def time_ordered_emission(
    nside: int,
    freq: float,
    components: List[Component],
    emissivities: List[float],
    line_of_sights: List[np.ndarray],
    observer_position: np.ndarray,
    earth_position: np.ndarray,
    pixel_chunk: np.ndarray,
    bin: bool,
    coord_out: str,
) -> np.ndarray:
    """Simulates and returns the Zodiacal emission timestream.

    Parameters
    ----------
    nside
        HEALPIX map resolution parameter.
    freq
        Frequency at which to evaluate the Zodiacal emission [Hz].
    components
        List of Zodiacal Components for which to simulate the emission.
    emissivities
        Sequency of emissivities at the frequency given by `freq`, one for each
        component.
    observer_position
        Heliocentric cartesian position of the observer.
    earth_position
        Heliocentric cartesian position of the Earth.
    bin
        If True, the time-ordered sequence of emission per pixel is binned 
        into a HEALPIX map. Defaults to False.
    coord_out
        Coordinate frame of the output map.

    Returns
    -------
        Simulated timestream of Zodiacal emission [MJy/sr] (optionally 
        binned into a HEALPIX map).
    """

    if bin:
        pixels, counts = np.unique(pixel_chunk, return_counts=True)
        unit_vectors = _get_unit_vectors(
            nside=nside,
            pixels=pixels,
            coord_out=coord_out,
        )
        emission = np.zeros((len(components), hp.nside2npix(nside)))

        for idx, component in enumerate(components):
            integrated_comp_emission = line_of_sight_integrate(
                radial_distances=line_of_sights[idx],
                get_emission_func=component.get_emission,
                observer_position=observer_position,
                earth_position=earth_position,
                unit_vectors=unit_vectors,
                freq=freq,
            )
            emission[idx, pixels] = integrated_comp_emission * emissivities[idx]

        emission[:, pixels] *= counts

        return emission * 1e20

    pixels, indicies = np.unique(pixel_chunk, return_inverse=True)
    unit_vectors = _get_unit_vectors(
        nside=nside,
        pixels=pixels,
        coord_out=coord_out,
    )
    time_stream = np.zeros((len(components), len(pixel_chunk)))

    for idx, component in enumerate(components):
        integrated_comp_emission = line_of_sight_integrate(
            radial_distances=line_of_sights[idx],
            get_emission_func=component.get_emission,
            observer_position=observer_position,
            earth_position=earth_position,
            unit_vectors=unit_vectors,
            freq=freq,
        )

        time_stream[idx] = integrated_comp_emission[indicies] * emissivities[idx]

    return time_stream * 1e20


def _get_unit_vectors(nside: int, pixels: np.ndarray, coord_out: str) -> np.ndarray:
    """Returns the unit vectors of a HEALPIX map given a requested output coordinate system.

    Since the Interplanetary Dust Model is evaluated in Ecliptic coordinates,
    we need to rotate any unit vectors defined in another coordinate frame to
    ecliptic before evaluating the model.
    """

    unit_vectors = np.asarray(hp.pix2vec(nside, pixels))
    if coord_out != "E":
        unit_vectors = hp.rotator.Rotator(coord=[coord_out, "E"])(unit_vectors)

    return np.asarray(unit_vectors)


from abc import ABC, abstractmethod
from dataclasses import dataclass
from math import radians, sin, cos
from math import pi as π
from typing import Tuple, Union

from numba import njit
import numpy as np

from zodipy._functions import interplanetary_temperature, blackbody_emission


@dataclass
class Component(ABC):
    """Base class for an Interplanetary Dust Component.

    This class defines a method for getting the coordinates to a shell at
    distance R around an observer in the primed coordinate system
    (component-centric ecliptic cartesian coordinates).

    Attributes
    ----------
    x_0
        x-offset from the Sun in helliocentric ecliptic cartesian coordinates.
    y_0
        y-offset from the Sun in helliocentric ecliptic cartesian coordinates.
    z_0
        z-offset from the Sun in helliocentric ecliptic cartesian coordinates.
    i
        Inclination with respect to the ecliptic plane [deg].
    Ω
        Ascending node [deg].
    """

    x_0: float
    y_0: float
    z_0: float
    i: float
    Ω: float

    def __post_init__(self) -> None:
        self.i = radians(self.i)
        self.Ω = radians(self.Ω)
        self.X_component = np.expand_dims([self.x_0, self.y_0, self.z_0], axis=1)

    @abstractmethod
    def get_density(
        self, R_prime: np.ndarray, Z_prime: np.ndarray, *, θ_prime: np.ndarray
    ) -> np.ndarray:
        """Returns the dust density at a shell around the observer.

        Parameters
        ----------
        R_prime
            Array of distances corresponding to discrete points along a
            line-of-sight for a shell surrounding the observer in the primed
            coordinates.
        Z_prime
            Heights above the midplane in primed coordinates of a component
            corresponding to the distances in `R_prime`.
        θ_prime
            Relative mean lognitude between the discrete points along the
            line-of-sight describe by `R_prime` and the Earth.

        Returns
        -------
            Density of the component at the coordinates given by R_prime,
            Z_prime, and θ_prime.
        """

    @staticmethod
    @njit
    def get_coordinates(
        R_comp: Union[float, np.ndarray],
        X_observer: np.ndarray,
        X_earth: np.ndarray,
        X_unit: np.ndarray,
        X_component: np.ndarray,
        Ω_component: float,
        i_component: float,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns a set of coordinates for a component.

        Given a set of heliocentric ecliptic positions in space given by the
        unit vectors `X_unit` and `R_comp`, we compute the `R_helio`, `R_prime`,
        `Z_prime`, and `θ_prime` coordinates as seen by and observer whos
        heliocentric ecliptic coordinates are given by `X_observer`. These are
        the coordinates required by the Interplanetary Dust Model to evalutate
        the density of a Zodiacal Component at the given positions.

        Parameters
        ----------
        R_comp
            Distance R to a shell centered on the observer at which we want
            to evaluate the Zodiacal emission at.
        X_observer
            Heliocentric ecliptic cartesian coordinates of the observer.
        X_earth
            Heliocentric ecliptic cartesian coordinates of the Earth.
        X_unit
            Heliocentric ecliptic cartesian unit vectors pointing to each 
            position in space we that we consider.
        X_component
            Heliocentric ecliptic cartesian off-set of the component
            (x_0, y_0, z_0).
        Ω_component
            Ascending node of the component.
        i_component
            Inclination of the component.

        Returns
        -------
        R_helio
            Array of distances corresponding to discrete points along a
            line-of-sight for a shell surrounding the observer in heliocentric
            ecliptic coordinates.
        R_prime
            Array of distances corresponding to discrete points along a
            line-of-sight for a shell surrounding the observer in the primed
            coordinates.
        Z_prime
            Heights above the midplane in primed coordinates of a component
            corresponding to the distances in `R_prime`.
        θ_prime
            Relative mean lognitude between the discrete points along the
            line-of-sight describe by `R_prime` and the Earth.
        """

        X_helio = R_comp * X_unit + X_observer
        R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)

        X_prime = X_helio - X_component
        R_prime = np.sqrt(X_prime[0] ** 2 + X_prime[1] ** 2 + X_prime[2] ** 2)

        Z_prime = (
            X_prime[0] * sin(Ω_component) * sin(i_component)
            - X_prime[1] * cos(Ω_component) * sin(i_component)
            + X_prime[2] * cos(i_component)
        )

        X_earth_prime = X_earth - X_component[0]
        θ_prime = np.arctan2(X_prime[1], X_prime[0]) - np.arctan2(
            X_earth_prime[1], X_earth_prime[0]
        )

        return R_helio, R_prime, Z_prime, θ_prime

    def get_emission(
        self,
        distance_to_shell: Union[float, np.ndarray],
        observer_position: np.ndarray,
        earth_position: np.ndarray,
        unit_vectors: np.ndarray,
        freq: float,
    ) -> np.ndarray:
        """Returns the emission at a shell of distance R from the observer.

        For a description on X_observer, X_earth, X_unit and R, please
        see the get_coords function.

        Parameters
        ----------
        distance_to_shell
            Distance R to a shell centered on the observer for which we want
            to evaluate the Zodiacal emission.
        observer_position
            Heliocentric ecliptic cartesian coordinates of the observer.
        earth_position
            Heliocentric ecliptic cartesian coordinates of the Earth.
        unit_vectors
            Heliocentric ecliptic cartesian unit vectors pointing to each 
            position in space we that we consider.
        freq
            Frequency at which to evaluate the Zodiacal emission.

        Returns
        -------
        emission
            Zodiacal emission at
            Array containing the Zodiacal emission emitted from a shell at
            distance R from the observer. The shape is (len(R), `NPIX`).
        """

        observer_position = np.expand_dims(observer_position, axis=1)
        R_helio, R_prime, Z_prime, θ_prime = self.get_coordinates(
            R_comp=distance_to_shell,
            X_observer=observer_position,
            X_earth=earth_position,
            X_unit=unit_vectors,
            X_component=self.X_component,
            Ω_component=self.Ω,
            i_component=self.i,
        )
        density = self.get_density(
            R_prime=R_prime,
            Z_prime=Z_prime,
            θ_prime=θ_prime,
        )
        temperature = interplanetary_temperature(R=R_helio)
        emission = blackbody_emission(T=temperature, ν=freq)

        return emission * density


@dataclass
class Cloud(Component):
    """The Zodiacal Diffuse Cloud component.

    This class represents the diffuse cloud in the K98 IPD model. It provides a
    method to estimate the density of the diffuse cloud at a shell around the
    observer.

    Attributes
    ----------
    n_0
        Density at 1 AU.
    α
        Radial power-law exponent.
    β
        Vertical shape parameter.
    γ
        Vertical power-law exponent.
    μ
        Widening parameter for the modified fan.
    """

    n_0: float
    α: float
    β: float
    γ: float
    μ: float

    def __post_init__(self) -> None:
        super().__post_init__()

    def get_density(self, R_prime: np.ndarray, Z_prime: np.ndarray, **_) -> np.ndarray:
        """See base class for documentation."""

        ζ = np.abs(Z_prime) / R_prime
        μ = self.μ
        g = np.zeros_like(ζ)

        condition = ζ < μ
        g[condition] = ζ[condition] ** 2 / (2 * μ)
        g[~condition] = ζ[~condition] - (μ / 2)

        return self.n_0 * R_prime ** -self.α * np.exp(-self.β * g * self.γ)


@dataclass
class Band(Component):
    """The Zodiacal Astroidal Band component.

    This class represents the Astroidal Dust Band components in the K98 IPD
    model. It provides a method to estimate the density of the dust bands at a
    shell around the observer.

    Attributes
    ----------
    n_0
        Density at 3 AU.
    δ_ζ
        Shape parameter [deg].
    v
        Shape parameter.
    p
        Shape parameter.
    δ_r
        Inner radial cutoff.
    """

    n_0: float
    δ_ζ: float
    v: float
    p: float
    δ_r: float

    def __post_init__(self) -> None:
        super().__post_init__()
        self.δ_ζ = radians(self.δ_ζ)

    def get_density(self, R_prime: np.ndarray, Z_prime: np.ndarray, **_) -> np.ndarray:
        """See base class for documentation."""

        ζ = np.abs(Z_prime) / R_prime
        ζ_over_δ_ζ = ζ / self.δ_ζ
        term1 = (3 * self.n_0 / R_prime) * np.exp(-((ζ_over_δ_ζ) ** 6))
        term2 = 1 + ((ζ_over_δ_ζ) ** self.p) / self.v
        term3 = 1 - np.exp(-((R_prime / self.δ_r) ** 20))

        return term1 * term2 * term3


@dataclass
class Ring(Component):
    """The Zodiacal Circum-solar Ring component.

    This class represents the Circum-solar Ring component in the K98 IPD model.
    It provides a method to estimate the density of the Circum-solar Ring at a
    shell around the observer.

    Attributes
    ----------
    n_0
        Density at 1 AU.
    R
        Radius of the peak density.
    σ_r
        Radial dispersion.
    σ_z
        Vertical dispersion.
    """

    n_0: float
    R: float
    σ_r: float
    σ_z: float

    def __post_init__(self) -> None:
        super().__post_init__()

    def get_density(self, R_prime: np.ndarray, Z_prime: np.ndarray, **_) -> np.ndarray:
        """See base class for documentation."""

        term1 = -(((R_prime - self.R) / self.σ_r) ** 2)
        term2 = np.abs(Z_prime) / self.σ_z

        return self.n_0 * np.exp(term1 - term2)


@dataclass
class Feature(Component):
    """The Zodiacal Earth-trailing Feature component.

    This class represents the Earth-trailing Feature component in the K98 IPD
    model. It provides a method to estimate the density of the Earth-trailing
    Feature at a shell around the observer.

    Attributes
    ----------
    n_0
        Density at 1 AU.
    R
        Radius of the peak density.
    σ_r
        Radial dispersion.
    σ_z
        Vertical dispersion.
    θ
        Longitude with respect to Earth.
    σ_θ
        Longitude dispersion.
    """

    n_0: float
    R: float
    σ_r: float
    σ_z: float
    θ: float
    σ_θ: float

    def __post_init__(self) -> None:
        super().__post_init__()
        self.θ = radians(self.θ)
        self.σ_θ = radians(self.σ_θ)

    def get_density(
        self, R_prime: np.ndarray, Z_prime: np.ndarray, θ_prime: np.ndarray
    ) -> np.ndarray:
        """See base class for documentation."""

        Δθ = θ_prime - self.θ

        condition1 = Δθ < -π
        condition2 = Δθ > π
        Δθ[condition1] = Δθ[condition1] + 2 * π
        Δθ[condition2] = Δθ[condition2] - 2 * π

        term1 = -(((R_prime - self.R) / self.σ_r) ** 2)
        term2 = np.abs(Z_prime) / self.σ_z
        term3 = (Δθ / self.σ_θ) ** 2

        return self.n_0 * np.exp(term1 - term2 - term3)



class ModelNotFoundError(Exception):
    """Raised if a pre-initialized IPD model is not found."""

class SimulationStrategyNotFoundError(Exception):
    """Raised if a non implemented simulation stategy is selected."""



from enum import Enum
from typing import Dict, Type

from zodipy._components import Component, Cloud, Band, Ring, Feature


class ComponentLabel(Enum):
    """Labels representing the six Zodiacal components in the K98 model."""

    CLOUD = "cloud"
    BAND1 = "band1"
    BAND2 = "band2"
    BAND3 = "band3"
    RING = "ring"
    FEATURE = "feature"


LABEL_TO_CLASS: Dict[ComponentLabel, Type[Component]] = {
    ComponentLabel.CLOUD: Cloud,
    ComponentLabel.BAND1: Band,
    ComponentLabel.BAND2: Band,
    ComponentLabel.BAND3: Band,
    ComponentLabel.RING: Ring,
    ComponentLabel.FEATURE: Feature,
}



import astropy.constants as const
import numpy as np

δ_K98 = 0.4668626
T_0 = 286

h = const.h.value
c = const.c.value
k_B = const.k_B.value


def blackbody_emission(T: np.ndarray, ν: float) -> np.ndarray:
    """Returns the blackbody emission for a temperature T and frequency ν.

    Parameters
    ----------
    T
        Temperature of the blackbody [K].
    ν
        Frequency [Hz].

    Returns
    -------
        Blackbody emission [W / m^2 Hz sr].
    """

    term1 = (2 * h * ν ** 3) / c ** 2
    term2 = np.expm1((h * ν) / (k_B * T))

    return term1 / term2


def interplanetary_temperature(
    R: np.ndarray,
    T_0: float = T_0,
    δ: float = δ_K98,
) -> np.ndarray:
    """Returns the Interplanetary Temperature given a radial distance from the Sun.

    Parameters
    ----------
    R
        Radial distance from the sun in ecliptic coordinates.
    T_0
        Temperature of the solar system at R = 1 AU.
    δ
        Powerlaw index.

    Returns
    -------
        Interplanetary temperature.
    """

    return T_0 * R ** -δ



from dataclasses import dataclass, field
from typing import Dict

import numpy as np

from zodipy._component_labels import ComponentLabel


EPS = np.finfo(float).eps  # Smallest non zero value.
RADIAL_CUTOFF = 6  # Distance to Jupiter in AU.


@dataclass
class IntegrationConfigRegistry:
    """Container for registered integration configs.
    
    By integration config, we mean a mapping of discrete points along a 
    line-of-sight per component.
    """

    registry: Dict[str, Dict[ComponentLabel, np.ndarray]] = field(default_factory=dict)

    def register_config(
        self,
        name: str,
        components: Dict[ComponentLabel, np.ndarray],
    ) -> None:
        """Adds a new integration config to the registry."""

        self.registry[name] = components

    def get_config(self, name: str) -> Dict[ComponentLabel, np.ndarray]:
        """Returns an integration config from the registry."""

        if name not in self.registry:
            raise ModuleNotFoundError(
                f"{name} is not a config in the registry. Avaliable configs are "
                f"{', '.join(self.registry)}"
            )
        return self.registry[name]


integration_config_registry = IntegrationConfigRegistry()

integration_config_registry.register_config(
    name="default",
    components={
        ComponentLabel.CLOUD: np.linspace(EPS, RADIAL_CUTOFF, 155),
        ComponentLabel.BAND1: np.linspace(EPS, RADIAL_CUTOFF, 50),
        ComponentLabel.BAND2: np.linspace(EPS, RADIAL_CUTOFF, 50),
        ComponentLabel.BAND3: np.linspace(EPS, RADIAL_CUTOFF, 50),
        ComponentLabel.RING: np.linspace(EPS, 2.25, 50),
        ComponentLabel.FEATURE: np.linspace(EPS, 1, 50),
    },
)

DEFAULT_INTEGRATION_CONFIG = integration_config_registry.get_config("default")




from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence

import numpy as np
import astropy.units as u

from zodipy._component_labels import ComponentLabel


@dataclass
class Emissivity:
    """Emissivity parameters for a Zodiacal component.

    Attributes
    ----------
    spectrum
        Frequencies (or wavelengths) corresponding to the emissivity fits.
    components
        Dictionary mapping components to the emissivity fits.
    """

    spectrum: u.Quantity
    components: Dict[ComponentLabel, Sequence[float]]

    def __call__(self, component: ComponentLabel, ν_or_λ: u.Quantity) -> float:
        """Returns the interpolated emissivity given a frequency.

        Parameters
        ----------
        component
            Component label.
        ν_or_λ
            Frequency or wavelength at which to to interpolate in the emissivities.

        Returns
        -------
        emissivity
            Interpolated emissivity scaling factor.
        """

        ν_or_λ = ν_or_λ.to(self.spectrum.unit, equivalencies=u.spectral())

        if not self.spectrum[0] <= ν_or_λ <= self.spectrum[-1]:
            raise ValueError(f"Frequency is out of range")

        emissivity = np.interp(ν_or_λ, self.spectrum, self.components[component])

        return emissivity


def get_emissivities(
    ν_or_λ: u.Quantity,
    emissivity: Optional[Emissivity],
    components: List[ComponentLabel],
) -> List[float]:
    """Returns a list of interpolated emissivities for each component."""

    if emissivity is not None:
        return [emissivity(component, ν_or_λ) for component in components]

    return [1.0 for _ in components]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zodipy']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.3.2,<9.0.0',
 'astropy>=4.3.1,<5.0.0',
 'astroquery>=0.4.3,<0.5.0',
 'h5py>=3.6.0,<4.0.0',
 'healpy>=1.15.0,<2.0.0',
 'numba>=0.54.1,<0.55.0',
 'numpy>=1.15',
 'scipy>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'zodipy',
    'version': '0.3.2',
    'description': 'Zodipy is a python tool that simulates the Zodiacal emission.',
    'long_description': '\n<img src="imgs/zodipy_logo.png" width="350">\n\n[![PyPI version](https://badge.fury.io/py/zodipy.svg)](https://badge.fury.io/py/zodipy)\n![Tests](https://github.com/MetinSa/zodipy/actions/workflows/tests.yml/badge.svg)\n[![astropy](http://img.shields.io/badge/powered%20by-AstroPy-orange.svg?style=flat)](http://www.astropy.org/)\n\n\n---\n\n\n*Zodipy* is a Python simulation tool for Zodiacal Emission (Interplanetary Dust Emission). It allows you to compute the \nsimulated emission in a timestream, or at an instant in time.\n\n![plot](imgs/zodi_default.png)\n\n## Installing\nZodipy is available at PyPI and can be installed with ``pip install zodipy``.\n\n## Features\nThe full set of features and use-cases will be documentated in the nearby future.\n\n**Initializing an Interplantery Dust Model:** We start by selecting which Interplanetary Dust Model to use. Currently, the implemented options are the [Kelsall et al. (1998)](https://ui.adsabs.harvard.edu/abs/1998ApJ...508...44K/abstract) model with or without the various emissivity fits from the Planck collaboration.\n```python\nimport zodipy\n\n# Other options for models are "K98", "Planck13", "Planck15"\nmodel = zodipy.InterplanetaryDustModel(model="Planck18")\n```\n\n**Instantaneous emission:** By obtaining the coordinates of an observer through the JPL Horizons API, we can simulate the full sky at an instant in time as follows:\n```python\nimport healpy as hp\n\nepoch = 59215  # 2010-01-01 in Modified Julian dates\nemission = model.get_instantaneous_emission(\n    nside=256, \n    freq=800, \n    observer="Planck", \n    epochs=epoch\n)\n\nhp.mollview(emission, norm="hist", coord=["E", "G"])\n```\n![plot](imgs/zodi_planck.png)\n\nThe `epochs` input must follow the convention used in [astroquery](https://astroquery.readthedocs.io/en/latest/jplhorizons/jplhorizons.html). If multiple dates are used as epochs, the returned emission will be the average emission over all instantaneous maps.\n\nAdditionally, it is possible to retrieve the emission of each Zodiacal Component in a dictionary, if the parameter `return_comps` is set to `True`. Following is an example of what each component may have looked like at 6th of October 2021.\n\n![plot](imgs/comps.png)\n\n\n**Time-ordered emission:** For a chunk of time-ordered data, it is possible to compute the simulated Zodiacal Emission over each observed pixel. In the following example we simulate the Zodiacal Emission time-stream given a chunk of the time-ordered pixels from the DIRBE instrument of the COBE satellite (Photometric Band 8, Detector A, first day of observations):\n```python\nimport matplotlib.pyplot as plt\nimport zodipy\n\nmodel = zodipy.InterplanetaryDustModel(model="K98")\n\ndirbe_pixel_timestream = ...    # Get in DIRBE tods\ndirbe_coords = ...  # Get coords of DIRBE at the time corresponding to the tod chunk \nearth_coords = ... # Get coords of the Earth at the time corresponding to the tod chunk \ndirbe_freq = 974    # GHz\ndirbe_nside = 128\n\ntimestream = model.get_time_ordered_emission(\n    nside=dirbe_nside,\n    freq=dirbe_freq,\n    pixels=dirbe_pixel_timestream,\n    observer_coordinates=dirbe_coords,\n    earth_coordinates=earth_coords\n)\n\nplt.plot(timestream)\n```\n![plot](imgs/timestream.png)\n\n\n**Binned time-ordered emission:** By setting the optional `bin` parameter to `True`, the emission is binned into a map which we can visualize as follows:\n\n```python\n\n# Get three tod chunks, each corresponding to a day of observation\npixel_chunks = [...]\ndirbe_coords = [...]\nearth_coords = [...]\n\n# Initialize empty emission and hit maps array\nemission = np.zeros(hp.nside2npix(nside))\nhits_map = np.zeros(hp.nside2npix(nside))   \n    \n    # Loop over tod chunks\n    for pixels, dirbe_coords, earth_coords in zip(pixel_chunks, dirbe_coords, earth_coords)):\n        \n        # We construct the total hits map over all chunks so that we can\n        # normalize the output map\n        unique_pixels, counts = np.unique(pixels, return_counts=True)\n        hits_map[unique_pixels] += counts\n\n        emission += model.get_time_ordered_emission(\n            freq=freq,\n            nside=nside,\n            pixels=pixels,\n            observer_coordinates=dirbe_coords,\n            earth_coordinates=earth_coords,\n            bin=True,\n        )\n\nemission /= hits_map\n\nhp.mollview(emission, norm="hist", coord=["E", "G"])\n```\n![plot](imgs/binned.png)\n',
    'author': 'Metin San',
    'author_email': 'metinisan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MetinSa/zodipy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from typing import Optional, Sequence, Union, Dict

from astroquery.jplhorizons import Horizons
import numpy as np


TARGET_MAPPINGS = {
    "planck": "Planck",
    "wmap": "WMAP",
    "l2": "32",
    "sun": "10",
    "mercury": "199",
    "venus": "299",
    "earth": "399",
    "mars": "499",
    "jupiter": "599",
}


def query_target_positions(
    target: str,
    epochs: Optional[Union[float, Sequence[float], Dict[str, str]]] = None,
) -> np.ndarray:
    """Returns the heliocentric cartesian positions of the target given some epoch.

    Parameters
    ----------
    target
        Name of the target body. Must be a valid target in the JPL Horizons API.
    epochs
        Either a list of epochs in JD or MJD format or a dictionary
        defining a range of times and dates; the range dictionary has to
        be of the form {``'start'``:'YYYY-MM-DD [HH:MM:SS]',
        ``'stop'``:'YYYY-MM-DD [HH:MM:SS]', ``'step'``:'n[y|d|h|m|s]'}.
        If no epochs are provided, the current time is used.

    Returns
    -------
        Heliocentric cartesian positions of the target.
    """

    if (target_lower := target.lower()) in TARGET_MAPPINGS:
        target = TARGET_MAPPINGS[target_lower]

    if epochs is None:
        epochs = 2459215.50000  # 01-01-2021

    query = Horizons(id=target, id_type="majorbody", location="c@sun", epochs=epochs)
    ephemerides = query.ephemerides()

    R = ephemerides["r"].value
    lon, lat = ephemerides["EclLon"].value, ephemerides["EclLat"].value
    lon, lat = np.deg2rad(lon), np.deg2rad(lat)

    x = R * np.cos(lat) * np.cos(lon)
    y = R * np.cos(lat) * np.sin(lon)
    z = R * np.sin(lat)

    return np.stack((x, y, z), axis=1)



from typing import Dict, Optional, Sequence, Union

import numpy as np
import astropy.units as u

from zodipy._astroquery import query_target_positions
from zodipy._emissivities import get_emissivities
from zodipy._integration_config import integration_config_registry
from zodipy._simulation import instantaneous_emission, time_ordered_emission
from zodipy.models import model_registry


class InterplanetaryDustModel:
    """The Zodipy simulation interface.

    The Interplanetary Dust Model used by Zodipy is the Kelsall et al. (1998)
    Interplanetary Dust Model, which includes five (six) Zodiacal components:
        - The Diffuse Cloud (cloud)
        - Three Asteroidal Bands (band1, band2, band3)
        - The Circumsolar Ring (ring) + The Earth-trailing Feature (feature)

    Optionally, it is possible to scale the K98 emission with component
    specific emissivities, as done by the Planck Collaboration. This is
    achieved by specifiying one of the following implemented models:
        - Planck13 (all five Kelsall components + emissivity fits for each)
        - Planck15 (cloud + bands + new emissivity fits)
        - Planck18 (cloud + bands + the latest emissivity fits)

    NOTE: No extrapolation is done when using the fitted emissivities, and as
    such, these models can only be evaluated within the frequency range covered
    by the Planck HFI Bands for which the emissivities were fitted.
    """

    def __init__(self, model: str = "K98") -> None:
        """Initializes the interface given an Interplanetary Dust Model.

        Parameters
        ----------
        model
            The name of the model to initialize.
        """

        self.model = model_registry.get_model(model)
        integration_config = integration_config_registry.get_config("default")
        self.line_of_sights = [
            line_of_sight for line_of_sight in integration_config.values()
        ]

    @u.quantity_input(freq_or_wavelength=("Hz", "m", "micron"))
    def get_instantaneous_emission(
        self,
        freq_or_wavelength: u.Quantity,
        nside: int,
        *,
        observer: str = "L2",
        epochs: Optional[Union[float, Sequence[float], Dict[str, str]]] = None,
        return_comps: bool = False,
        coord_out: str = "E",
    ) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        """Simulates and returns the instantaneous Zodiacal Emission [MJy/sr].

        By instantaneous emission we mean the emission observed at an instant
        in time. If multiple epochs are given, the returned emission will be
        the mean of all simulated instantaneous observations.

        The observer location, given by the parameter `observer` (and
        optionally the location of the Earth if either of the Feature or the
        Ring components are included in the selected Interplanetary Dust Model)
        are queried from the Horizons JPL ephemerides, given some epoch defined
        by the `epochs` parameter.

        NOTE: This function returns the fullsky emission from at a single time.
        This means that we in the simulation evaluate line-of-sights that
        sometimes points directly towards the inner Solar System and through
        the Sun, where the dust density increases exponentially. Such
        line-of-sights are unlikely to be observed by an actual observer, and
        as such, the simulated emission will appear very bright in these
        regions.

        Parameters
        ----------
        freq_or_wavelength
            Frequency or wavelength at which to evaluate the Zodiacal emission.
        nside
            HEALPIX map resolution parameter of the returned emission map.
        observer
            The name of the observer for which we quiery its location given
            the `epochs` parameter. Defaults to 'L2'.
        epochs
            The observeration times given as a single epoch, or a list of epochs
            in JD or MJD format, or a dictionary defining a range of times and
            dates; the range dictionary has to be of the form
            {'start':'YYYY-MM-DD [HH:MM:SS]', 'stop':'YYYY-MM-DD [HH:MM:SS]',
            'step':'n[y|d|h|m|s]'}. If no epochs are provided, the current time
            is used in UTC.
        return_comps
            If True, the emission is returned component-wise in a dictionary.
            Defaults to False.
        coord_out
            Coordinate frame of the output map. Defaults to 'E' (heliocentered
            ecliptic coordinates).

        Returns
        -------
        emission
            Simulated (mean) instantaneous Zodiacal emission [MJy/sr].
        """

        observer_positions = query_target_positions(observer, epochs)
        if self.model.includes_earth_neighboring_components:
            earth_positions = query_target_positions("earth", epochs)
        else:
            earth_positions = observer_positions.copy()

        emissivities = get_emissivities(
            ν_or_λ=freq_or_wavelength,
            emissivity=self.model.emissivities,
            components=list(self.model.components.keys()),
        )

        freq = freq_or_wavelength.to("Hz", equivalencies=u.spectral()).value
        emission = instantaneous_emission(
            nside=nside,
            freq=freq,
            components=list(self.model.components.values()),
            emissivities=emissivities,
            observer_positions=observer_positions,
            earth_positions=earth_positions,
            line_of_sights=self.line_of_sights,
            coord_out=coord_out,
        )

        if return_comps:
            return {
                comp.value: emission[idx]
                for idx, comp in enumerate(self.model.components)
            }

        return emission.sum(axis=0)

    @u.quantity_input(freq_or_wavelength=("Hz", "m", "micron"))
    def get_time_ordered_emission(
        self,
        freq_or_wavelength: u.Quantity,
        nside: int,
        *,
        pixels: np.ndarray,
        observer_position: np.ndarray,
        earth_position: Optional[np.ndarray] = None,
        bin: bool = False,
        return_comps: bool = False,
        coord_out: str = "E",
    ) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        """Simulates and returns the Zodiacal emission [MJy/sr] in a timestream.

        Given a sequence of time-ordered pixels, the Zodiacal emission is
        evaluated from a constant location in space given by the
        `observer_position`. The `earth_position` is required for Interplanetary
        Dust models that include the Earth-trailing Feature and Circum-solar
        Ring components.

        Parameters
        ----------
        freq_or_wavelength
            Frequency or wavelength at which to evaluate the Zodiacal emission.
        nside
            HEALPIX map resolution parameter of the returned emission map.
        pixels
            Sequence of time-ordered pixels.
        observer_position
            The heliocentric ecliptic cartesian position of the observer at 
            the time of observing the tods.
        earth_position
            The heliocentric ecliptic cartesian position of the Earth at the 
            time of observing the tods. If None, the observer is assumed to be 
            the Earth. Defaults to None.
        bin
            If True, the time-ordered sequence of emission per pixel is binned
            into a HEALPIX map. Defaults to False.
        return_comps
            If True, the emission is returned component-wise in a dictionary.
            Defaults to False.
        coord_out
            Coordinate frame of the output map. Defaults to 'E' (heliocentric 
            ecliptic coordinates).

        Returns
        -------
        emission
            Simulated timestream of Zodiacal emission [MJy/sr] (optionally
            binned into a HEALPIX map).
        """

        if earth_position is None:
            earth_position = observer_position

        emissivities = get_emissivities(
            ν_or_λ=freq_or_wavelength,
            emissivity=self.model.emissivities,
            components=list(self.model.components.keys()),
        )

        freq = freq_or_wavelength.to("Hz", equivalencies=u.spectral()).value
        emission = time_ordered_emission(
            nside=nside,
            freq=freq,
            components=list(self.model.components.values()),
            emissivities=emissivities,
            line_of_sights=self.line_of_sights,
            observer_position=observer_position,
            earth_position=earth_position,
            pixel_chunk=pixels,
            bin=bin,
            coord_out=coord_out,
        )

        if return_comps:
            return {
                comp.value: emission[idx]
                for idx, comp in enumerate(self.model.components)
            }

        return emission.sum(axis=0)

    def __str__(self) -> str:
        """String representation of the InterplanetaryDustModel."""

        reprs = []
        for label, component in self.model.components.items():
            component_repr = f"{component.__class__.__name__}" + "\n"
            reprs.append(f"({label.value}): {component_repr}")

        main_repr = "InterplanetaryDustModel("
        main_repr += f"\n  name: {self.model.name}"
        main_repr += "\n  components( "
        main_repr += "\n    " + "    ".join(reprs)
        main_repr += "  )"
        main_repr += "\n)"

        return main_repr